// 																																			KEDS Project
// 																												(Kansas Event Data System)
//__________________________________________________________________________________
//	Left-Right-Left Hidden Markov model for international event sequences 	HMM.LRL 

/*  This program is designed to estimate and fit WEIS event sequences using a 
	  left-right-left Hidden Markov model.

	  Modifications by Philip A. Schrodt (p-schrodt@ukans.edu)
	  Date: 97.3.10.
	    
	  1. Program is based on C++ code originally written by Richard Myers 
	     and James Whitson;see the header file for detailed copyright and
	     modification history.

	  2. Files were translated into Metrowerks C (Macintosh).  This involved moving
	     from the object model of C++ to a bunch of clunky globals and combining
	     the various original programs.  Less elegant but it works...

		3. Modified the input scheme to work with BCOW sequences -- which are rather
			 long -- and WEIS codes.  

		4. The functions under "continuous WEIS sequences" do a running computation of
			 the alphas for a given .hmm matrix by month, doing a fit of the previous 
			 100 days in each case.  This produces a tab-delimited file. 
			 
		COMMENTS ON CURRENT CODE <97.8.12>
		This thing got built quite quickly in order to get estimates for the APSA 
		paper, and there is a *lot* of idiosyncratic control via #if statements and
		commenting-out.  It all makes sense given the tasks in the paper but at the
		moment is this *not a general program* -- instead it is full of idiosyncracies
		that perform specific tasks required for the analysis. 
		
		At the moment these include:
		1. Program generally writes over files; the assumption is that the names of
		   earlier versions of the files will have been changed if they need to be
		   saved.
		   
		2. Some of the routines do not include the 1982-1985 period; others do.
		
		3. qprob at the moment looks at two consecutive states, not just one.
		
		4. The file-structure documentation below is generally out of date.
		
		5. The cross-correlation function XCORR has a quirky if() that gets around
		   a splice in the data set.
		   
		6. There are a couple of places where loops could be moved inside if() 
		   statements to speed up the program.
		   
		7. System stops iterating on the estimation after four alpha-reversals
		 
		I also pulled out an assortment of debugging and diagnostic code before posting 
		this to the web site; hopefully this did not cause any fatal damage.  
*/     
//____________________     
/*	HMM MODELS
		Line 1		Number of states
		Line 2		Number of symbols
		remainder	pp	pr	pn	s1  s2  ...  sn
							pp = prob of transition to previous state
							pr = prob of recurrence 
							pp = prob of transition to next state 
							s1, s2,...,sn = symbol probabilities 
*/
//____________________     
/*	OPTIONS:
		At present, control of the functions of the program are managed by #ifs and 
		commenting-out in main().
*/
//____________________
/*	Programmer: Philip A. Schrodt
		Compiler:		Metrowerks C
		Language:		ANSI C with // comments

		Revision History: 
			97.2		Original modification of Meyers & Whitson program
	    97.6		Modification to handle random starting points; Viterbi algorithm
	    97.8		Modification to estimate various LRL models; date-based templates;
	    				cross-correlation optimization
*/
//__________________________________________________________________________________

#include <assert.h>
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include "hmmlrl.h"

//__________________________________________________________________________________
//																																			main program

void main(void)
{
char  sTarget[] = "LEB";
char  sSource[] = "ISR";

char template_dates[] = "WEIS.templ.dates";

int			ka, nex,seq_length ;
double r;

time_t tStart,tEnd;
	
	Initialize();
	tStart = time(NULL);
	
 
#if TRUE				// code for estimating HMM using templates

 	Read_Dyadic_WEIS(sSource, sTarget);
	Init_Template_HMM(template_dates);
	Init_Saved_HMM(0);
  dump_model("HMM.ISRLEBDY.LRL");
	Compute_Running_Prob (sSource, sTarget,'L');

#endif


#if FALSE				// code for computing alpha and qprob from a prebuilt model

 	Read_Dyadic_WEIS(sSource, sTarget);
	Init_Prebuilt_HMM("HMM.ISRLEB.P77");
	Compute_Running_Prob (sSource, sTarget,'L');

#endif

#if FALSE				// code for optimizing predictive models

 	Read_Indicator();
 	Read_Dyadic_WEIS(sSource, sTarget);
	Init_Template_HMM(template_dates); 
	for (nex = 0; nex < XC_EXPER; nex++)
	{ 
		printf("\nXC Exper %3d\n",nex);
		Estimate_HMM();
		Init_Saved_HMM(0);					// get the best model
		Compute_Running_Prob (sSource, sTarget,'R');
	  r = Sum_XCorr();
		check_save_XC(r);
    for (ka=0; ka < MAX_SAVE; ka++) printf("%2d   %10.2f\n",ka,aSaveXC[ka]);  // debugging
	  printf("\n");
	}
	Init_Saved_HMMXC(0);
	dump_model("HMM.ISRLEBDY.XCP.0");
	Init_Saved_HMMXC(1);
	dump_model("HMM.ISRLEBDY.XCP.1");

#endif

	tEnd = time(NULL);
	printf("\a\nFinished at %s\n",ctime(&tEnd));  
	printf("Elapsed time %8.2f minutes\n",(difftime(tEnd,tStart)/60.0));
} // main

//__________________________________________________________________________________
//																						main programs from Meyers and Whitson

void Train_HMM(int seq_length)
// create a hidden markov model from a series of symbol sequences using a randomly 
// generated model.
{
double 	min_delta_psum = MIN_DELTA; 

char newfilename[100];	
int  ka;  
    
  Init_Random_HMM();    											// initialize model randomly
  batch_train(seq_length,min_delta_psum);    // train model on data
  																						// dump the resulting model to a file
//  sprintf(newfilename,"%s.train.hmm",train_file);
//  dump_model(newfilename);
	  
} // Train_HMM

void Test_HMM(char *initial_model, char *train_file)
// Computes the log-alpha fit for a set of sequences	  
{																				
	Init_Prebuilt_HMM(initial_model);  		// initialize model from model file
  batch_test(train_file,MAX_SAVE);  							// test model on data
} // Test_HMM

void Test_Saved_HMM(int isave, char *train_file)
// Computes the log-alpha fit for a set of sequences	  
{																				
	Init_Saved_HMM(isave);  		// initialize model from model file
  batch_test(train_file,isave);  							// test model on data
} // Test_HMM


//__________________________________________________________________________________
//																																				utilities

void Initialize (void )
//
{
int 	ka, seed;

	srand(time(NULL));					// seed the random number generator

	for (seed=0; seed < MAX_SAVE; seed++) 
	{
		aSaveAlpha[seed] = 1.0;   // flag to show cell does not contain real data
		aSaveXC[seed] = 100.0;
	}
	
} // Initialize

double UnifRnd( void )
// Uniform[0,1] random number generator
{
return  ((double)rand())/RAND_MAX;
}

//__________________________________________________________________________________
//																																		HMM procedures

void Init_Prebuilt_HMM(char *filename)
// set max_states, max_symbols and model parameters from file
{
	printf("\nUsing predefined model %s\n",filename);
	load_model(filename);
//	show_probs();
} // Init_Prebuilt_HMM

void Init_Saved_HMM(int isave)
// set max_states, max_symbols and model parameters from file
{
int s;

	printf("\nUsing saved model %2d\n",isave);
  Set_Size();
     for (s=0; s < max_states; s++) 
				aStates[s] = aSaveSt[s][isave];
//	show_probs();
} // Init_Saved_HMM

void Init_Saved_HMMXC(int isave)
// set max_states, max_symbols and model parameters from file
{
int s;

	printf("\nUsing saved model %2d\n",isave);
  Set_Size();
     for (s=0; s < max_states; s++) 
				aStates[s] = aSaveStXC[s][isave];
} // Init_Saved_HMMXC

void Set_Size(void)
// set globals max_symbols and max_states
{
  max_symbols = RUN_SYMBOLS;
  max_states = RUN_STATES;
} //Set_Size

void Init_Random_HMM(void)
// create hmm, initialize model parameters with random values.
{
  Set_Size(); 						// set size globals

  rnd_init_hmm_probs();  // rnd initialize the trans and emit probs

//  printf("\nUsing a random model...\n");
//  show_probs();
} //Init_Random_HMM



void batch_train( int seq_length, double min_delta_psum) 
// train hmm from the existing sequence file until the sum of the probability
// changes is less than min_delta_psum.

// *** Changes by Emdad Khan: the code for the unequal length strings
// is changed to batch mode. It does not work in nonbatch mode.
// The equations given in Rabiner and elsewhere also indicates batch 
// mode training. Yi[] and yi[][] are used to add the batch mode update

// *** Changes in hmm.pa: sequences are now read a single time in a main();
//     seq_length is now passed instead of a file name
{
#define REVERSE_THRESHOLD 3

  int str;

  // local arrays and integers used in the case with strings of unequal lengths
  arState 			y1, y2, y3, y4, y6; 
  arStateSymb		y5;
  int i,j,k;
  int kreverse = 0;							// check number of alpha-reverse conditions, exit if >REVERSE_THRESHOLD

  double prob_sum=999999.0;						 // insure this will get reset
  double alpha_tot,last_alpha_tot;

  trellis_width = seq_length+1;  							// define trellis dimensions

																										  // loop until local min in probs
  for (str=0,alpha_tot=0.0; str<num_strings; str++)
    alpha_tot+=test(strings[str],string_len[str]);
#if fWRITE_FILE
  printf("\nlog(alpha)\tprob_sum\n");
  printf("----------\t--------\n");
  printf("%f\n",alpha_tot);
#endif
																					  // do batch update for each case:
  while (prob_sum > min_delta_psum) 
  {
    last_alpha_tot = alpha_tot;
    for (i=0 ; i < max_states; i++) 
    {
      y1[i] = 0;
      y2[i] = 0;
      y3[i] = 0;
      y4[i] = 0;
			y6[i] = 0;

			for (k=0; k < max_symbols; k++) 
			{
			  y5[i][k] = 0;
			}
    }

	  for (str=0; str<num_strings; str++) 
	  {																							// sum estimates across all training cases
			set_ab_counts(strings[str],string_len[str]);

			for (i=0; i < max_states; i++)
			{
			  y1[i] += a_numer_sum_recur[i];
		    y2[i] += a_numer_sum_next[i]; 
			  y3[i] += a_numer_sum_prev[i]; 
			  y4[i] += a_denom_sum[i]; 
			  y6[i] += b_denom_sum[i];

			  for (k=0; k < max_symbols; k++) 
			  {
			    y5[i][k] += b_numer_sum[i][k]; 
			  }
			}
   	}  // end of for str= loop

    // Transferring data to proper arrays/variables to be used 
    // in the reestimate routine
    for (i=0; i < max_states; i++) 
    {
			a_numer_sum_recur[i] = y1[i];
			a_numer_sum_next[i]  = y2[i]; 
			a_numer_sum_prev[i]	 = y3[i]; 
			a_denom_sum[i]			 = y4[i]; 
			b_denom_sum[i] 			 = y6[i];

			for (k=0; k < max_symbols; k++) 
			{
			  b_numer_sum[i][k] = y5[i][k]; 
			}
    }

    // restimate routine initializes the prob_sum
    prob_sum = reestimate();

    for (str=0, alpha_tot=0.0; str<num_strings; str++) 
				alpha_tot+=test(strings[str],string_len[str]);     

#if fWRITE_FILE
  printf("%f\t%f\n",alpha_tot,prob_sum);				// running output of alpha values
#endif
  if (alpha_tot < last_alpha_tot) 
  {
#if fWRITE_FILE
    printf("\tWARNING: alpha values reversed!\n"); // should not occur! (but it does!!...)
#endif
    kreverse++;
    if (kreverse > REVERSE_THRESHOLD) break;   // bail out if this occurs too often
   }
 } // while 

//  printf("\nDone training model %s  ...\n",filename);
  curAlpha = alpha_tot;
#if fWRITE_FILE
  printf("alpha_tot = %10.2f\n",alpha_tot);
#endif
  // show_probs();

} //batch_train 

double batch_test(char *filename, int isave)
// test hmm from sequence file.
{
int 		seq_length;
int 		i, s, k;
double	total=0.0, val;
FILE		*ofp; 						// output file to write alpha values
char 		tempname[100];	

  sprintf(tempname,"LRL.circ.w.%s.%d",filename,isave);  // dump the resulting sequences to filename.seq
  ofp = fopen(tempname, "w");
  if (!ofp)
  {
  	printf("\aCan't open %s; skipping batch_test\n",tempname);
  	return(0);
  }
																									// write current model to ofp
  fprintf(ofp,"tot_alpha = %10.2f\n\n",aSaveAlpha[isave]);
  for ( s=0; s < max_states; s++) 
  {
    fprintf(ofp,"%8.4f\t",set_prev_trans(&aStates[s],-1.0));
    fprintf(ofp,"%8.4f\t",set_recur_trans(&aStates[s],-1.0));
    fprintf(ofp,"%8.4f\t",set_next_trans(&aStates[s],-1.0));
    for ( k=0; k < max_symbols; k++)
      fprintf(ofp,"%8.4f\t",set_symb_out(&aStates[s],k,-1.0));
    fprintf(ofp,"\n\n");
  }
  fprintf(ofp,"\n");

  seq_length = Read_Sequences(filename);  // load string_matrix and set globals

  trellis_width = seq_length+1;  // define trellis dimensions

  // loop through all test strings
  printf("\ntest on strings...\n");
  for (i=0; i<num_strings; i++) 
  {
    val = test(strings[i],string_len[i]);
    printf("alpha for %24.24s =  %10.4f ",string_name[i],val);
    printf( " len= %3d",string_len[i]);
    printf("\n");
    fprintf(ofp, "%24.24s\t %10.4f\n",string_name[i], val); // writing alpha values to file
    total+=val;
 
    Viterbi(i);
  }
  fclose(ofp);
  printf("------------ \nbatch alpha = %10.4f\n",total);

  printf("\nDone testing model on %s...\n",filename); 
return(total);
}

void check_save(void)
// check whether we should save the model.
{
int ki, k, s;

  for (k=0; k < MAX_SAVE; k++) 
  {
    if ((aSaveAlpha[k] > 0.0) | (curAlpha > aSaveAlpha[k]))
    {	
     	printf("               Improved curAlpha = %10.2f\n",curAlpha);

		 	if (k != MAX_SAVE-1)									// move remaining values down one slot
	    { for (ki=MAX_SAVE-1; ki > k; ki--) 
		    { aSaveAlpha[ki] =  aSaveAlpha[ki-1];
		      for (s=0; s < max_states; s++) 
						aSaveSt[s][ki] = aSaveSt[s][ki-1];
				}
			}
																						// put new information in slot k
		 aSaveAlpha[k] = curAlpha;
     for (s=0; s < max_states; s++) 
				aSaveSt[s][k] = aStates[s];
		break;																	// we're done		
		}
  }
} // check_save(void)

void check_save_XC(double sumpr)
// check whether we should save the XC model.
{
int ki, k, s;

  for (k=0; k < MAX_SAVE; k++) 
  {
    if ((aSaveXC[k] > 10.0) | (sumpr > aSaveXC[k]))
    {	
     	printf("               Improved sumpr = %10.2f\n",sumpr);

		 	if (k != MAX_SAVE-1)									// move remaining values down one slot
	    { for (ki=MAX_SAVE-1; ki > k; ki--) 
		    { aSaveXC[ki] =  aSaveXC[ki-1];
		      for (s=0; s < max_states; s++) 
						aSaveStXC[s][ki] = aSaveStXC[s][ki-1];
				}
			}
																						// put new information in slot k
		 aSaveXC[k] = sumpr;
     for (s=0; s < max_states; s++) 
				aSaveStXC[s][k] = aStates[s];
		break;																	// we're done		
		}
  }
} // check_save_XC(void)

void show_probs(void)
// output the model parameters.
{
int k,s;

  printf("\nstates: %d\nsymbols: %d\n",max_states,max_symbols);
  for (s=0; s < max_states; s++) 
  {
    printf("%8.4f ",set_prev_trans(&aStates[s],-1.0));
    printf("%8.4f ",set_recur_trans(&aStates[s],-1.0));
    printf("%8.4f ",set_next_trans(&aStates[s],-1.0));
    for (k=0; k < max_symbols; k++)
      printf("%8.4f ",set_symb_out(&aStates[s],k,-1.0));
    printf("\n\n");
  }
  printf("\n");
} // show_probs(void)

void dump_model(char *fname)
// output the model parameters to a file.
{
int 		k,s;
FILE 		*fout;

	fout=fopen(fname,"w");
  if (!fout) 
  {
    printf("ERROR: Couldn't create file for model dump.");
    exit(-1);
  }

  fprintf(fout,"states: %d\n",max_states);
  fprintf(fout,"symbols:  %d\n",max_symbols);
  for ( s=0; s < max_states; s++) 
  {
    fprintf(fout,"%8.4f\t",set_prev_trans(&aStates[s],-1.0));
    fprintf(fout,"%8.4f\t",set_recur_trans(&aStates[s],-1.0));
    fprintf(fout,"%8.4f\t",set_next_trans(&aStates[s],-1.0));
    for ( k=0; k < max_symbols; k++)
      fprintf(fout,"%8.4f\t",set_symb_out(&aStates[s],k,-1.0));
    fprintf(fout,"\n\n");
  }
  fprintf(fout,"\n");
  fclose(fout);
  printf("\nDumped model to file ==> %s\n",fname);
}

#if FALSE
void dump_seq(FILE *seq, int length)
// generate a random sequence according to model parameters.
{
int 		j, cur_state,k,symbol;
double 	prob_trans,prob_out,accum;

  cur_state=0;
  for (j=0; j < length; j++) 
  {
    																										// fill in random
    prob_trans = UnifRnd();    
    prob_out = UnifRnd();    
    if (prob_trans < set_recur_trans(&aStates[cur_state],-1.0))
    {
      accum = 0.0;
      symbol=max_symbols-1;
      for (k=0; k < max_symbols; k++) 
      {
				if (prob_out < set_recur_out(&aStates[cur_state],k,-1.0) + accum) 
				{
				  symbol = k;
				  break;
				}
				else
				  accum += set_recur_out(&aStates[cur_state],k,-1.0);
			}
      fprintf(seq,"%2d ",symbol);
    }
    else 
    {
      accum = 0;
      symbol=max_symbols-1;
      for ( k=0; k < max_symbols; k++) 
      {
				if (prob_out < set_next_out(&aStates[cur_state],k,-1.0) + accum) 
				{
				  symbol = k;
				  break;
				}
				else
				  accum += set_next_out(&aStates[cur_state],k,-1.0);
			}
	    fprintf(seq,"%2d ",symbol);
	    cur_state++;
    }
  }

  if (cur_state < max_states-1)  printf("WARNING: Sequence did not reach final state!\n");
  fprintf(seq,"\n");

} // dump_seq
#endif

//__________________________________________________________________________________
//																															model initialization

#if fLR

void rnd_init_hmm_probs(void)
// randomly initialize the model parameters for LR model
{
double 	pr;
arSymb	prob_syms;
int 		i,j;
																				// set aij
  for (i=1; i < max_states-1; i++)
  {
    pr = UnifRnd();    
    set_recur_trans(&aStates[i],pr);
    set_next_trans(&aStates[i],1-pr);
	}
																				// set terminal state
  i = max_states-1;
  set_recur_trans(&aStates[i],1.0);
  set_next_trans(&aStates[i],0.0); 			// no next transition

  for (i=0; i < max_states; i++)			// set b-ik
  {
    rnd_sym_probs(prob_syms);
    for (j=0; j < max_symbols; j++)
    {
			set_symb_out(&aStates[i],j,prob_syms[j]);
		}
  }

} // rnd_init_hmm_probs for LR

#else

void rnd_init_hmm_probs(void)
// randomly initialize the model parameters
{
double 	pr, pp, pn, prob_sum;
arSymb	prob_syms;
int 		i,j;
																				// set aij
	if (fCIRC)
	{
	  for (i=0; i < max_states; i++)
	  {
	    pp = UnifRnd();    
	    pr = UnifRnd();    
	    pn = UnifRnd();
	    prob_sum = pp + pr + pn;    
	    set_prev_trans(&aStates[i],pp/prob_sum);
	    set_recur_trans(&aStates[i],pr/prob_sum);
	    set_next_trans(&aStates[i],pn/prob_sum);
		}
	}
	else 
	{
	  for (i=1; i < max_states-1; i++)
	  {
	    pp = UnifRnd();    
	    pr = UnifRnd();    
	    pn = UnifRnd();
	    prob_sum = pp + pr + pn;    
	    set_prev_trans(&aStates[i],pp/prob_sum);
	    set_recur_trans(&aStates[i],pr/prob_sum);
	    set_next_trans(&aStates[i],pn/prob_sum);
		}
																					// set state 0
	  pr = UnifRnd();    
	  set_prev_trans(&aStates[0],0.0); 			// no prev transition
	  set_recur_trans(&aStates[0],pr);
	  set_next_trans(&aStates[0],1.0-pr);
																					// set terminal state
	  i = max_states-1;
	  pr = UnifRnd();    
	  set_prev_trans(&aStates[i],1.0 - pr);
	  set_recur_trans(&aStates[i],pr);
	  set_next_trans(&aStates[i],0.0); 			// no next transition
	}																	

  for (i=0; i < max_states; i++)			// set b-ik
  {
    rnd_sym_probs(prob_syms);
    for (j=0; j < max_symbols; j++)
    {
			set_symb_out(&aStates[i],j,prob_syms[j]);
		}
  }

} // rnd_init_hmm_probs
#endif

void rnd_sym_probs(double prob_syms[])
// set output symbol probabilities.
{
double 	prob_sum = 0.0;
int 		s;

  for (s=0; s < max_symbols; s++) 
  {
    prob_syms[s] = UnifRnd();
    prob_sum += prob_syms[s];
  }

  for (s=0; s < max_symbols; s++)  // normalize so they add to one
    prob_syms[s] = prob_syms[s] / prob_sum;

} // rnd_sym_probs

//__________________________________________________________________________________
//																																	file operations

void load_model(char *filename)
// set model parameters from a file.
{
FILE 		*param_file;
double 	prob;
int 		i,j;
	
  param_file = fopen(filename,"r");
  if (!param_file) { printf("ERROR: can not open %s\n",filename); exit(-1); }
  if (fscanf(param_file,"states: %d\n",&max_states)!=1) { printf("ERROR: Problem reading states: field of %s\n",filename); exit(-1);}
  if (fscanf(param_file,"symbols: %d\n",&max_symbols)!=1 ) { printf("ERROR: Problem reading symbols: field of %s\n",filename); exit(-1);}

  for (i=0; i < max_states; i++) 
  {
    fscanf(param_file,"%lf ",&prob);
    set_prev_trans(&aStates[i],prob);
    fscanf(param_file,"%lf ",&prob);
    set_recur_trans(&aStates[i],prob);
    fscanf(param_file,"%lf ",&prob);
    set_next_trans(&aStates[i],prob);

    for (j=0; j < max_symbols; j++) 
    {
      fscanf(param_file,"%lf ",&prob);
      set_symb_out(&aStates[i],j,prob);
    }
    fscanf(param_file,"\n");

  }
  fclose(param_file);
} // load_model


int Read_Sequences(char* filename)
// figure out number of symbols and number of sequences in file
// and load strings matrix and string_len array.
// returns the length of the longest sequence string (seq_length)
{
  int 	str,sym,tmp,val,i, seq_length;
  FILE 	*fin;
  char 	sline[MAX_LINE];

  for (i=0; i<MAX_STRINGS; i++) 
  {
    for (sym=0; sym<MAX_LENGTH; sym++) // initialize array to -1
      strings[i][sym]=-1;
  }

  fin = fopen(filename,"r");  					// open file

  seq_length= 0;
  num_strings = 0;
  str=0;
  fgets(sline,MAX_LINE-1,fin);			// read first line
  while (sline[0] == 'b') 
		if (strchr(sline,'*'))			// check for a skip indicator
		{
	   	do															// skip the sequence
	   		fgets(sline,MAX_LINE-1,fin);
	    while (sline[0] <= '9');
		}
		else							// read it
		{ 
			i=strlen(sline);								// save the sequence name
			sline[i-1]='\0';								// (cancel the \n)
			strcpy(string_name[str],sline);

			num_strings++;
			sym=-1;					 
	   	do															// read a sequence
	   	{	sym++;
	   		fgets(sline,MAX_LINE-1,fin);
	   		if (sline[0] <= '9') 					// read a symbol
	   		{
	   			sscanf(sline,"%*s %d",&val);
	   			strings[str][sym] = val;
				}
	    }
	    while (sline[0] <= '9');
	    
	    string_len[str++]=sym;
			if (seq_length < sym)
	      seq_length = sym;      
		}

  printf("\nmax sequence length = %d\n",seq_length);
  printf("number of strings = %d\n",num_strings);
  max_symbols = MAX_SYMBOLS;

  fclose(fin);

  return seq_length;
} // Read_Sequences

//__________________________________________________________________________________
//																																			calculations

void rescale_alphas(int col)
// rescale alpha values (from Rabiner).
{ 
int i;

  scaling_factors[col] = 0.0;
  for ( i=0; i < max_states; i++) 
  {
    scaling_factors[col] += alpha[col][i];
  }

  for (i=0; i <  max_states; i++)        // rescale all the alpha's and smooth them
  {
    alpha[col][i] /= scaling_factors[col];
  }
} // rescale_alphas

void rescale_betas(int col)
// rescale beta values after rescaling alphas.
{ 
int i;

  for ( i=0; i < max_states; i++)  			// rescale all the beta's w alpha's factors
  {
    beta[col][i] /= scaling_factors[col];
  }
} // rescale_betas

double alpha_F(int* symbol_array, int seq_length)
// calculate alpha trellis and return (and set) final sequence probability
{ 
int 		t,j;
int 		m = max_states-1;
double 	acirc = 0.0;					// extra term for circular model

  seq_length=(seq_length<0?trellis_width-1:seq_length);

  // clear the trellis and set the first column
  for (t=0; t < trellis_width; t++)
    for (j=0; j < max_states; j++) 
    {
      alpha[t][j] = 0.0;
    }
  alpha[0][0] = 1.0;  // first state starts with alpha = 1.0
  rescale_alphas(0);  // first col rescale is div by one

  // calculate the alphas for rest of cols
  for (t = 0; t < seq_length; t++) 
  { 		
    if (fCIRC) 
    	acirc = alpha[t][m] * set_next_trans(&aStates[m],-1.0);
    	
    alpha[t+1][0] = (
								      (alpha[t][0] * set_recur_trans(&aStates[0],-1.0) ) +
											(alpha[t][1] * set_prev_trans(&aStates[1],-1.0) )  +
											acirc
									  ) *
					  				set_symb_out(&aStates[0],symbol_array[t],-1.0);

    for (j=1; j < m; j++) 
    {
      alpha[t+1][j] = (
									      (alpha[t][j] * set_recur_trans(&aStates[j],-1.0) ) +
												(alpha[t][j-1] * set_next_trans(&aStates[j-1],-1.0) ) +
												(alpha[t][j+1] * set_prev_trans(&aStates[j+1],-1.0) )
										  ) *
						  				set_symb_out(&aStates[j],symbol_array[t],-1.0);
    }

    if (fCIRC) 
    	acirc = alpha[t][0] * set_prev_trans(&aStates[0],-1.0);

    alpha[t+1][m] = (
								      (alpha[t][m] * set_recur_trans(&aStates[m],-1.0) ) +
											(alpha[t][m-1] * set_next_trans(&aStates[m-1],-1.0) ) +
											acirc
									  ) *
					  				set_symb_out(&aStates[m],symbol_array[t],-1.0);
   rescale_alphas(t+1);
  }
										// compute probability of complete sequence
  prob_seq = 0.0;
  for (j=1; j < m; j++)
  	prob_seq += alpha[seq_length][j];
  	
  return (prob_seq);
} // alpha_F

double beta_I(int* symbol_array, int seq_length)
// calculate beta trellis and return initial beta value
{
int 		t,j;
int 		m = max_states-1;
double 	acirc = 0.0;					// extra term for circular model
	
 seq_length=(seq_length<0?trellis_width-1:seq_length);     

																							 // clear the trellis and set the last column
 for ( t=0; t < trellis_width; t++)
   for ( j=0; j < max_states; j++) 
   {
		 beta[t][j] = 0.0;
   }

	for ( j=0; j < max_states; j++) 
	{
		beta[seq_length][j]	 = 1.0;  		// final betas are 1.0 
	}

 																							// begin setting all the cols except last one
 for ( t = seq_length-1; t >= 0; t--) 
 {
    if (fCIRC) 
    	acirc =  ( beta[t+1][m] * set_prev_trans(&aStates[m],-1.0) *
				         set_symb_out(&aStates[m],symbol_array[t],-1.0)	);

	  beta[t][0] = ( 
		 								beta[t+1][0] * set_recur_trans(&aStates[0],-1.0) *
				            set_symb_out(&aStates[0],symbol_array[t],-1.0)
				          ) + (
				     				beta[t+1][1] * set_next_trans(&aStates[0],-1.0) * 
						     	  set_symb_out(&aStates[1],symbol_array[t],-1.0) 
				          ) + acirc; 
				          
   for ( j=1; j < m; j++) 
   {
		 beta[t][j] = ( 
		 								beta[t+1][j] * set_recur_trans(&aStates[j],-1.0) *
				            set_symb_out(&aStates[j],symbol_array[t],-1.0)
				          ) + (
				     				beta[t+1][j+1] * set_next_trans(&aStates[j],-1.0) * 
						     	  set_symb_out(&aStates[j+1],symbol_array[t],-1.0) 
 				          ) + (
				     				beta[t+1][j-1] * set_prev_trans(&aStates[j],-1.0) * 
						     	  set_symb_out(&aStates[j-1],symbol_array[t],-1.0)
						     	); 
  }

  if (fCIRC) 
  	acirc =  ( beta[t+1][0] * set_next_trans(&aStates[m],-1.0) *
			         set_symb_out(&aStates[0],symbol_array[t],-1.0)	);

	 beta[t][m] = ( 
	 								beta[t+1][m] * set_recur_trans(&aStates[m],-1.0) *
			            set_symb_out(&aStates[m],symbol_array[t],-1.0)
			          ) + (
			     				beta[t+1][m-1] * set_prev_trans(&aStates[m],-1.0) * 
					     	  set_symb_out(&aStates[m-1],symbol_array[t],-1.0)
				        ) + acirc; 

   rescale_betas(t);
 }

 return (beta[0][0]);
} // beta_I


void compute_gamxi(int* symbol_array, int seq_length)
// compute xi and gamma values and fill tables for entire trellis.
//
{
int i,t,j;
int 	m = max_states-1;
double accum;
 
  seq_length=(seq_length<0?trellis_width-1:seq_length);

 																		 	// clear xis
  for ( t=0; t < seq_length; t++ )
    for ( j=0; j < max_states; j++ ) 
    {
			xi_prev[t][j] = 0.0;
			xi_recur[t][j] = 0.0;
			xi_next[t][j] = 0.0;
			gammaF[t][j] = 0.0;
    }

 																			 // calc the xi and gamma tables
  for ( t = 0; t < seq_length; t++ ) 
  {
    accum = 0;
    for ( i=0; i < max_states; i++ ) 
    {
      if (i>0)
      { xi_prev[t][i] = ( alpha[t][i] *
			  									set_prev_trans(&aStates[i],-1.0) *
			  									set_symb_out(&aStates[i-1],symbol_array[t],-1.0) *
			  									beta[t+1][i-1] );
			 }
			 else if (fCIRC)
	      { xi_prev[t][0] = ( alpha[t][0] *
				  									set_prev_trans(&aStates[i],-1.0) *
				  									set_symb_out(&aStates[m],symbol_array[t],-1.0) *
				  									beta[t+1][m] );
			 }
			 else xi_prev[t][0] = 0.0;

      if (i<m)
      {
	      xi_next[t][i] = ( alpha[t][i] *
			  									set_next_trans(&aStates[i],-1.0) *
			  									set_symb_out(&aStates[i+1],symbol_array[t],-1.0) *
			  									beta[t+1][i+1] );
			}
			else if (fCIRC)
      { xi_next[t][m] = ( alpha[t][m] *
			  									set_next_trans(&aStates[m],-1.0) *
			  									set_symb_out(&aStates[0],symbol_array[t],-1.0) *
			  									beta[t+1][0] );
			}
			else xi_next[t][m] = 0.0;
			
      xi_recur[t][i] = (	alpha[t][i] * 
		   										set_recur_trans(&aStates[i],-1.0) * 
		  									 	set_symb_out(&aStates[i],symbol_array[t],-1.0) *
		   										beta[t+1][i] );
				   									
		  gammaF[t][i] = xi_prev[t][i] + xi_recur[t][i] + xi_next[t][i];
		  accum +=  gammaF[t][i];
    } // for i

    for ( i=0; i < max_states; i++ )    // now adjust for the sum
		{
		 xi_prev[t][i] /= accum;
		 xi_recur[t][i] /= accum;
		 xi_next[t][i] /= accum;
		 gammaF[t][i] /= accum;
		}
  }  // for t
    
} // compute_gamxi


double a_numer(int i, int j, int seq_length)
// this is the numerator of the a_ij reestimate and denom of b_ij reestimate,
// this func assumes gammas have been calc'd.
// ### NOTE: reverse the loop/if once this is working
{
double sum = 0.0;
int t;

  seq_length=(seq_length<0?trellis_width-1:seq_length);

  if ((j<0) || (j>=max_states))
  { if (fCIRC)
  	{	if ( j < 0 )
	 		{
			  assert(i == 0);		// ###
			  for( t=0; t < seq_length-1; t++)
			  	 sum += xi_prev[t][i];
	 		}
	 		else 
	 		{
			  assert(i == max_states-1); // ###
			  for( t=0; t < seq_length-1; t++)
			  	 sum += xi_next[t][i];
	 		}
  	}
  	return sum;
	} 


  for( t=0; t < seq_length-1; t++)   // -1 is consistent with Rabiner
    {
      if (i==j) 
      {
				sum += xi_recur[t][i];
      }
      else if ( j == i+1) 
      {
				sum += xi_next[t][i];
      }
      else if ( j == i-1) 
      {
				sum += xi_prev[t][i];
      } else 
      {
				printf("WARNING: xi_next[t][%d] shouldn't be requested\n",i);
				break;
      }
    }
  return sum;
} // a_numer


double a_denom(int i, int seq_length)
// this is the denominator of the a_ij reestimate.
{ 
int t;
double sum = 0.0;

  seq_length=(seq_length<0?trellis_width-1:seq_length);
	
  for( t=0; t < seq_length-1; t++)
      sum += gammaF[t][i];

  return sum;
} // a_denom

// 
double b_numer(int i, int sym, int *symbol_array, int seq_length)
// this is the numerator of the b_i,sym reestimate.
{	
int t;
double sum = 0.0;

  seq_length=(seq_length<0?trellis_width-1:seq_length);
	
  for( t=0; t < seq_length; t++) 
  {
   if ( symbol_array[t] == sym ) 
    {
			 sum += gammaF[t][i];
    }
  }
return sum;
} // b_numer

double b_denom(int i, int seq_length)
// this is the denominator of the b_i,sym reestimate.
{ 
int t;
double sum = 0.0;

  seq_length=(seq_length<0?trellis_width-1:seq_length);
	
  for( t=0; t < seq_length; t++)
      sum += gammaF[t][i];

  return sum;
} // b_denom


double set_ab_counts(int symbol_array[], int seq_length)
// set ab count from one string.
{
double 	alpha_tot;
int 		i,k;
double prob_ck=0;

  alpha_tot = alpha_F(symbol_array,seq_length);		// fill the alpha matrix
  
  beta_I(symbol_array,seq_length);  							// fill the beta matrix

  compute_gamxi(symbol_array,seq_length);  				// fill the xi and gamma matrices

  for (i=0; i < max_states; i++) 
  {
	  a_numer_sum_prev[i] = a_numer(i,i-1,seq_length);
	  a_numer_sum_recur[i] = a_numer(i,i,seq_length);
	  a_numer_sum_next[i] = a_numer(i,i+1,seq_length);

	  a_denom_sum[i] = a_denom(i,seq_length);
	  b_denom_sum[i] = b_denom(i,seq_length);

	  for ( k=0; k < max_symbols; k++) 
	  {
	    b_numer_sum[i][k] = b_numer(i,k,symbol_array,seq_length);
	  }
  }

  return(alpha_tot);
} // set_ab_counts

double reestimate()
// reestimate parameters after calculating ab_counts
{
  double prob, diff_sum = 0.0;
	int i,k,j;
	double check_prob;
	
  // loop all the transition probs

  for ( i=0; i < max_states; i++)
  {
																										    // do all the prob transitions except last one
    prob = a_numer_sum_recur[i] / a_denom_sum[i];
    diff_sum += fabs(prob - set_recur_trans(&aStates[i],-1.0));
    set_recur_trans(&aStates[i],prob);
 																											 // do all the next transitions
    prob = a_numer_sum_next[i] / a_denom_sum[i];
    diff_sum += fabs(prob - set_next_trans(&aStates[i],-1.0));
    set_next_trans(&aStates[i],prob);

 																											// do all the prev transitions
    prob = a_numer_sum_prev[i] / a_denom_sum[i];
    diff_sum += fabs(prob - set_prev_trans(&aStates[i],-1.0));
    set_prev_trans(&aStates[i],prob);
																												// compute observation probs
    for ( k=0; k < max_symbols; k++) 
    {
			prob = b_numer_sum[i][k] / b_denom_sum[i];
			diff_sum += fabs(prob -	set_symb_out(&aStates[i],k,-1.0));
			set_symb_out(&aStates[i],k,prob);
    }
  }

  return diff_sum;
} // reestimate()

//
double test(int* string, int seq_length)
// test model for a single string.
{
  double log_alpha;
	int t;
	
  seq_length=(seq_length<0?trellis_width-1:seq_length);

  // fill alpha trellis and scaling coeffs
  log_alpha=log(alpha_F(string,seq_length));

  // calc log probability from coeffs
  if (scaling_factors[0]>0.0) 
  { // if scaling_factors set
    log_alpha=0.0;
    for ( t=0; t < seq_length+1; t++)
      log_alpha += log(scaling_factors[t]);
  }

	Viterbi(0);						// use only if computing qstar

  return log_alpha;
} // test

//__________________________________________________________________________________
//																																	aState functions

double set_symb_out(state *pState, int symbol, double prob)
// set/return the probability of generating a particular symbol for pState, 
// (smooth with MIN_PROB)
{
  if (prob != -1.0) 
  {
    pState->symb_out[symbol] = prob;
  }
  if (pState->symb_out[symbol] < MIN_PROB)
    pState->symb_out[symbol] = MIN_PROB;
  return pState->symb_out[symbol];
} // set_symb_out

double set_recur_trans(state *pState,double prob)
// set/return the probability of making a recurrent transition.
// (smooth with MIN_PROB)
{
  if (prob != -1.0) 
  {
    pState->recur_trans = prob;
  }
  if (pState->recur_trans < MIN_PROB)
    pState->recur_trans = MIN_PROB;
  return pState->recur_trans;
} // set_recur_trans

double set_next_trans(state *pState,double prob)
// set/return the probability of making transitioning to the next state.
// (smooth with MIN_PROB)
{
  if (prob != -1.0) 
  {
    pState->next_trans = prob;
  }
  if (pState->next_trans < MIN_PROB)
    pState->next_trans = MIN_PROB;
  return pState->next_trans;
} // set_next_trans

double set_prev_trans(state *pState,double prob)
// set/return the probability of making transitioning to the next state.
// (smooth with MIN_PROB)
{
 #if fLR
    pState->prev_trans = 0;
#else
 if (prob != -1.0) 
  {
    pState->prev_trans = prob;
  }
  if (pState->prev_trans < MIN_PROB)
    pState->prev_trans = MIN_PROB;
#endif
  return pState->prev_trans;
} // set_prev_trans

//__________________________________________________________________________________
//																																Viterbi algorithm
// 																							(compute most likely state sequence)

void Viterbi( int index)
// compute MLE state sequence for string[index] using	Viterbi algorithm from Rabiner
// Note: this does not do any rescaling of the deltas, so on a long sequence these
//       could underflow.

{
int 		j, kb, kt;
double 	recur_del, next_del, prev_del,dmax;			// values of delta for probs
int			maxi;											// state that produced max psi 

// Initialization
	for (j=0; j < max_states; j++)
	{
		delta[0][j] = set_symb_out(&aStates[j],strings[index][0],-1.0) ;
		psi[0][j] = 0;
	}
	
// fill trellis using recursive formula 
	for (kt=1; kt < string_len[index]; kt++)
	{
		for (j=0; j < max_states; j++)			// compute delta and psi for each state
		{
			recur_del = delta[kt-1][j]*set_recur_trans(&aStates[j],-1.0) ;
			if (j>0)
				next_del = delta[kt-1][j-1]*set_next_trans(&aStates[j-1],-1.0) ;
			else 
				next_del = 0;
			if (j<max_states - 1)
				prev_del = delta[kt-1][j+1]*set_prev_trans(&aStates[j+1],-1.0) ;
			else 
				prev_del = 0;
																				// find maximum
			if (recur_del > next_del)
			{
				dmax = recur_del;
				maxi = j;
			}
			else 
			{
				dmax = next_del;
				maxi = j-1;
			}
			if (prev_del > dmax)
			{
				dmax = prev_del;
				maxi = j+1;
			}
		
			delta[kt][j] = dmax * set_symb_out(&aStates[j],strings[index][kt],-1.0) ;
			psi[kt][j] = maxi;
		}		
	}			// kt

// backtrack through psi to get the optimal sequence

	kt = string_len[index]-1;
	maxi=0;
	recur_del = delta[kt][0];							// use recur_del to store maximum
	
	for (j=1; j < max_states; j++)			// find maximum delta[T]
		if (recur_del < delta[kt][j])
		{
			recur_del = delta[kt][j];
			maxi = j;
		}
	qstar[kt] = maxi;
	
	for (kt=string_len[index]-1; kt > 0 ; kt--)
	{
		qstar[kt-1] = psi[kt][qstar[kt]];		
	}
} // Viterbi

double qprop(int state, int length)
// return proportion of qstar that have the value state
// Note: this is currently summing over a set of two states
{
int ka, kount=0;

	for (ka= 0; ka<length; ka++)
//		if (qstar[ka] == state) kount++;
		if ((qstar[ka] == state) || (qstar[ka] == state+1)) kount++;
	return ((double)kount/length);
} // qprop


//__________________________________________________________________________________
//																												Continuous WEIS comparsion

long Julian (char *s)
// Converts dates of form YYMMDD to Julian serial beginning 0 = 1 Jan 1904
// (which matches the Excel convention) 
{
int 	mo,dy;
long	yr, ij;

  for (mo=0; mo<7; mo++) if (s[mo] == ' ') s[mo] = '0';			// convert blanks to zeros
  
																		//Break date out of the string and convert to integers 
	yr = (short)s[0] * 10 + (short)s[1] - 532;			// 532 substracts the 1904 base from yr
	mo = (short)s[2] * 10 + (short)s[3] - 528;
	dy = (short)s[4] * 10 + (short)s[5] - 528;
						
	ij =  yr * 365;
	ij += ((yr+3) / 4) + Month[mo] + (dy-1);				// ((yr+3)/4) is the leap-year adjustment
	if (((yr % 4) == 0) && (mo > 2)) ij++;	// adjust for month within a leap-year
	return ij;
} // Julian

void Read_Ex_WEIS(char *SourceCode, char *TargetCode)
// read a WEIS ME file, excluding events that do not have BCOW equivalents
{
FILE 	*fin;
int		ka=0,kb;
char  fname[32];
int		exclude[]={6,7,10,11,13,14,15,16,17,20,0}; // 0 is the end-of-list marker
long	date;
int		evt;

	printf("\nReading sequence for %3.3s > %3.3s",SourceCode,TargetCode);

	sprintf(fname,"WEISME %3.3s>%3.3s",SourceCode,TargetCode);
	fin = fopen(fname,"r");
	if (!fin) { printf("Error: can't open %s",fname); exit(0); }
//	printf("======================\n");

	while ((!feof(fin)) && (ka<MAX_WEIS))
	{
		fscanf(fin,"%d %d",&date,&evt);
		kb=0;
		while ((exclude[kb]) && (exclude[kb++] != evt)) {};  // check if evt is in the exclusion list
		if (!(exclude[kb]))
		{  aWEIS_Date[ka] = date;
			 aWEIS_Evt[ka++] = evt;
		}
	}
	
	if (ka >= MAX_WEIS) { printf("Error: file length exceeds MAX_WEIS in %s",fname); exit(0); }

	numWEIS=ka;	
	fclose(fin);
} // Read_Ex_WEIS

void Read_Dyadic_WEIS(char *SourceCode, char *TargetCode)
// read a WEISDY file into aWEIS arrays
{
FILE 	*fin;
int		ka = 0;
char  fname[32];
long	date;
int		evt;

	printf("\nReading sequence for %3.3s > %3.3s",SourceCode,TargetCode);

	sprintf(fname,"WEISDY.%3.3s>%3.3s",SourceCode,TargetCode);
	fin = fopen(fname,"r");
	if (!fin) { printf("Error: can't open %s",fname); exit(0); }
//	printf("======================\n");

	while ((!feof(fin)) && (ka<MAX_WEIS))
	{
		fscanf(fin,"%d %d",&date,&evt);
		aWEIS_Date[ka] = date;
		aWEIS_Evt[ka++] = evt;
	}
	
	if (ka >= MAX_WEIS) { printf("Error: file length exceeds MAX_WEIS in %s",fname); exit(0); }

	numWEIS=ka;	
	fclose(fin);
} // Read_Dyadic_WEIS

void Init_Template_HMM(char *sFile)
// Reads a set of dates from sFile; moves SEQ_SIZE events into strings, 
// then estimates a model
{
FILE 	*fin;
int		ka = 0, kb;
char 	sdate[8];		// leave room for the \n (and then some...)
int		lastDate;
long	ijul;
int		kstring=0;
int		nex;

	printf("\nReading templates from ",sFile);
	fin = fopen(sFile,"r");
	if (!fin) { printf("Error: can't open %s",sFile); exit(0); }

	fgets(sdate,8,fin);
	while (sdate[0] != '=')
	{		
		ijul = Julian(sdate);
		lastDate=SEQ_SIZE-1;
		while (( aWEIS_Date[++lastDate] < ijul) && (lastDate < numWEIS)) {};  // find the date in the array

		kb=0;
		for (ka = lastDate - SEQ_SIZE; ka < lastDate; ka++)			// load previous SEQ_SIZE events into strings
			strings[kstring][kb++] = aWEIS_Evt[ka];
		string_len[kstring] = SEQ_SIZE;
		fgets(sdate,8,fin);
		kstring++;
	}
	fclose(fin);
	num_strings = kstring;

// NOTE: The  Estimate() function could be used in here...
  Set_Size();
	for (nex = 0; nex < N_EXPER; nex++)
	{ printf("\nExper %3d\n",nex);
		Train_HMM(SEQ_SIZE);
	  check_save();
    for (ka=0; ka < MAX_SAVE; ka++) printf("%2d   %10.2f\n",ka,aSaveAlpha[ka]);  // debugging
	  printf("\n");
	}
	printf("\nModel estimation finished\n");

} // Init_Template_HMM

void Estimate_HMM(void)
// MC estimation from existing set of strings
{
int nex, ka;

	for (nex=0; nex < MAX_SAVE; nex++) aSaveAlpha[nex] = 1.0;   // flag to show cell does not contain real data

  Set_Size();
	for (nex = 0; nex < N_EXPER; nex++)
	{ 
		Train_HMM(SEQ_SIZE);
	  check_save();

		#if fWRITE_FILE
		printf("\nExper %3d\n",nex);
    for (ka=0; ka < MAX_SAVE; ka++) printf("%2d   %10.2f\n",ka,aSaveAlpha[ka]);  // debugging
	  printf("\n");
		#endif

	}
	#if fWRITE_FILE
	printf("\nModel estimation finished\n");
	#endif
} // Estimate_HMM

void Next_Month (char *sMonth)
// increment the month string
{
  sMonth[3]++;
  if (sMonth[3] > '9')
  { sMonth[3] = '0';
  	sMonth[2] = '1';
  }
  else if ((sMonth[3] == '3') && (sMonth[2] == '1'))
  { sMonth[3] = '1';
  	sMonth[2] = '0';
	  sMonth[1]++;
	  if (sMonth[1] > '9')
	  { sMonth[1] = '0';
	  	sMonth[0]++;
	  }
	}  
} // Next_Month

void Compute_Running_Prob (char *SourceCode, char *TargetCode, char ctype )
// computes alpha for 100 events prior to the end of each month
{
char 		sStart[] = "790801";
char 		sEnd[]	 = "970701";						// final month plus 1
int 		ka, kb, ijul; 
int			lastDate=SEQ_SIZE-1;
double	val, val2 = 0.0;
int			kpr = 0;

#if fWRITE_FILE
FILE		*fout; 												// output file to write alpha values
char  	fname[32];

	printf("\nComputing alphas for %3.3s > %3.3s\n",SourceCode,TargetCode);

	sprintf(fname,"WEIS.%3.3s>%3.3s.%c.alphas",SourceCode,TargetCode,ctype);
	fout = fopen(fname,"w");
	if (!fout) { printf("Error: can't open %s",fname); exit(0); }
#endif

// set globals
  trellis_width = SEQ_SIZE+1;  // define trellis dimensions
  max_symbols = MAX_SYMBOLS;
  string_len[0] = SEQ_SIZE;
	
	while (strcmp(sStart,sEnd))
	{
		ijul = Julian(sStart);
		while (( aWEIS_Date[++lastDate] < ijul) && (lastDate < numWEIS)) {};  // find the date in the array

		kb=0;
		for (ka = lastDate - SEQ_SIZE; ka < lastDate; ka++)			// load previous MAX_SEQ events into strings
			strings[0][kb++] = aWEIS_Evt[ka];

    val = test(strings[0],SEQ_SIZE);												// compute alpha
    val2 = qprop(1,SEQ_SIZE);															// compute % of state i and i+1

#if fWRITE_FILE
    printf("alpha for %s =  %10.4f  qprop =  %10.4f\n",sStart,val,val2);
    fprintf(fout, "%s\t %10.4f\t %10.4f\n",sStart, val,val2); // write values to file
#endif
    
    if ( (ijul < 28490) || (ijul>29950))   // fill prstate, skipping 1982-1985
    	prstate[kpr++] = val2;

		Next_Month(sStart);
	}	

#if fWRITE_FILE
  fclose(fout);
#endif

} // Compute_Running_Prob


double XCorr(double x[], double y[], int xa, int ya, int max)
// cross-correlation of x[xa...] x y[ya...].  Upper limit is whichever 
// index is first >=max
{
#define iSPLICE 28    // location of the 1981-1986 splice in the file

double sx = 0, sxa = 0, sy = 0, sya = 0, ssr = 0;
int na=0;				// sample size
	
	while( (xa < max) && (ya < max)) 
	{	
		if (((xa>iSPLICE) && (ya>iSPLICE)) ||			// don't sum across the splice in the data set
			 ((xa<=iSPLICE) && (ya<=iSPLICE)))
		{
	   sx += x[xa];
	   sy += y[ya];
	   sxa += x[xa] * x[xa];
	   sya += y[ya] * y[ya];
	   ssr += x[xa] * y[ya];
	   na++;
		}
		xa++;
		ya++;
	}
 	return (ssr - (sx*sy/na))/sqrt((sxa - (sx * sx)/na)*(sya-(sy * sy)/na));
} // XCorr

void Read_Indicator(void)
// read the indicator file
{
#define sINDIC_NAME "indicat.tft"

FILE 	*fin;
int	date;

	printf("\nReading indicator sequence %s\n",sINDIC_NAME);

	fin = fopen(sINDIC_NAME,"r");
	if (!fin) { printf("Error: can't open %s\n",sINDIC_NAME); exit(0); }

	nMonth = 0;
	while ((!feof(fin)) && (nMonth<MAX_MONTHS))
	{
		fscanf(fin,"%*d %d",&date);
		indic[nMonth++] = date;
	}
	
	if (nMonth >= MAX_MONTHS) 
	{ printf("Error: file length exceeds MAX_MONTH in %s\n",sINDIC_NAME); exit(0); }

	fclose(fin);
} // Read_Indicator

double Sum_XCorr(void)
// compute the sum of the target cross-correlations
{
int ka;
int nLim = nMonth;
double r, sum=0;

//	nLim = 87;							// split sample tests
	
	for (ka = 2; ka<5; ka++)				// do lagged values of prstate
	{
		sum += XCorr(prstate,indic,0,ka,nLim);
	}
	
	return(sum);
} // Sum_XCorr


