// 																																			KEDS Project
// 																												(Kansas Event Data System)
//__________________________________________________________________________________
// 																	hmmlrl: left-right-left Hidden Markov model for 
// 														 							international event sequences
//__________________________________________________________________________________
// 
//
// ORIGINAL COPYRIGHT INFORMATION
//   file : hmm.cc
// version: 1.03 [August 21, 1995]
/* Copyright (C) 1994 Richard Myers and James Whitson
   Permission is granted to any individual or institution to use, copy, or
   redistribute this software so long as all of the original files are 
   included unmodified, that it is not sold for profit, and that this 
   copyright notice is retained.
*/

/*  *************************************************************
    Updated by Emdad Khan (Computer & Information Science dept,
    U.C Santa Cruz and National Semiconductor, Santa Clara, CA).
    Date 5/12/95. Brief description of the changes follows:

    (1)
    The original model written by Richard & James seemed to works for equal
    length string only. For unequal length strings (i.e vq coded words),
    i.e when different people say the same words, the program does not 
    converge (basically oscillates). The authors used "instanteneous"
    updates for each strings for the unequal case. Such scheme nullifies
    the learning a string by another etc. The commonly used approach
    is the "batch" update: a good reference is Rabiner's book or paper
    (as alos referenced by the authors). 
     
    So, I updated the code (training routine and other associated routines)
    to use batch update for unequal length strings. More detailed descrip-
    tion of the changes are given in corresponding routines i.e
    batch_train, beta_I, compute_gamma, alpha_F and reestimate routines.

    (2) 
    Some of equations related to training (namely, gamma, beta, alpha)
    showed discrepencies compared to the equations given in HMM literatures.
    Those equations are corrected and details of the changes are shown in
    corresponding routines. These routines are basically same as the 
    routines mentioned above.

    (3) 
    Some of the parameters were initialised both in the main routines as 
    well as in the function declarations in the .h file. Some of the 
    compilers does not like this. So, initialization of parameters are
    changed so that it is present only in one place. 
    
    *******************************************************************
*/ 
 
/*  *************************************************************
    Updated by Arthur Stephens (Arthur.Stephens@src.bae.co.uk)
    Date 8/30/95.

    Purified code, plugged memory leaks, corrected some out of bounds
    array conditions and got it to compile on Solaris. (Thanks Art --rm)
*/

/*  *************************************************************
    Modifications by Philip A. Schrodt (p-schrodt@ukans.edu)
    Date: 97.3.10.

    1. Files were translated into Metrowerks C (Macintosh).  This involved moving
       from the object model of C++ to a bunch of clunky globals and combining
       most of the original files.  Less elegant but it works...
       
    Modifications for hmm.pa.
    Date: 97.6.22.
    
    1. Code for doing the tests against BCOW data removed.
    
    2. N_EXPER different initial values were used and the best MAX_SAVE of these
       are saved and written to a file.
       
    3. Viterbi algorithm and qprop calculations were added.
       
    Modifications for hmm.lrl.
    Date: 97.7/8.

    1. Added code to implement the Viterbi algorithm for generating the most
       probable state sequence (from Rabiner 1989) 
       
    2. Added a transition to previous state; eliminated the transition symbol
    	 vectors.  fCIRC controls whether chain is circular.   Rewrote the estimation 
    	 routines to match Rabiner.   I/O formats changed to handle the LRL parameters.
    	 
    	 Also added the fLR switch that forces the model to be LR, albeit with the
    	 new estimation code rather than M&W's.
    	 
    3. Eliminated the equal-length case code;
    
    4. Added a system for estimating cases based on the SEQ_SIZE events prior to a 
       set of dates read from a file.
       
    5. Added system for handling dyadic behaviors by concatenating the obverse
       dyads code with an increment of 22 to the original dyad; most of the code
       and files for this involve a DY suffix somewhere.
    	 
    6. Added another Monte-Carlo layer to look for a maximum involving the cross-
       correlation ("xcorr") of the percentage of time in various state probabilities
       and an external indicator (e.g. TFT measure or Goldstein; this is read from
       a file.
       
    7. Continued to control all of these options with a rat's-nest of #if statements... 
    
*/

#define MAX_TRELLIS		  120		// determined by maximum sequence length
#define MAX_SYMBOLS			 45		// dyadic WEIS sequences
#define MAX_STATES			  6
#define MAX_STRINGS			 16		// maximum number of training strings
#define MAX_LENGTH		  120		// max length of training strings
#define	NAME_LENGTH			 32
#define	MAX_SAVE					4		// number of experiments saved
#define	MIN_DELTA				1.0		// min_delta_psum  (originally 0.01)
#define	MAX_WEIS			12500		// maximum aWEIS size
#define SEQ_SIZE 				100		// prior sequence length

#define	N_EXPER				 8		// number of experiments with random initial matrices

															// values for the globals max_symbols and max_states
#define RUN_SYMBOLS			 45
#define RUN_STATES			  6

#define	XC_EXPER			 6		// number of XC experiments
#define XC_SIZE					 12		// maximum xcorr lag/lead computed
#define XC_ARRAY					2*XC_SIZE + 1
#define MAX_MONTHS 			215		// max length of monthly data in indic. file

// some global constants

const double MIN_PROB = 0.000001;	// never let alpha or beta equal zero
const int 	 MAX_LINE = 40;				// length of input lines

// flags
#define fCIRC			 	FALSE				// circular matrix?
#define fLR					TRUE				// LR matrix?
#define fWRITE_FILE	TRUE				// Set FALSE when optimizing prediction; turns off writing


//_____________________________________________________________________________
// Typedefs

typedef double arTrellState[MAX_TRELLIS][MAX_STATES] ;	
typedef double arTrell[MAX_TRELLIS] ;	
typedef double arState[MAX_STATES] ;	
typedef double arStateSymb[MAX_STATES][MAX_SYMBOLS] ;	
typedef double arNumSymb[MAX_SYMBOLS];	
typedef double arSymb[MAX_SYMBOLS];	

typedef struct 
{
	arNumSymb 	symb_out; 				// array of weights for output symbols
  double 			prev_trans;
  double 			recur_trans;
  double 			next_trans;
} state; 

//___________________________________________________________________________
// Prototypes

//	------- Meyers and Whitson main programs -------
void Train_HMM(int seq_length);
void Test_HMM(char *initial_model, char *train_file);

//	------- modified main programs -------
void Test_Saved_HMM(int isave, char *train_file);			// Computes the log-alpha fit for a set of sequences	  

//	------- Utilities -------
void Initialize (void );
double UnifRnd( void );						// uniform [0,1] random number generator

//	------- aSTATE FUNCTIONS -------
double set_symb_out(state *aState,int symbol, double prob);
double set_recur_trans(state *aState,double prob);
double set_next_trans(state *aState, double prob);
double set_prev_trans(state *aState, double prob);

//	------- MODEL INIT -------
void rnd_init_hmm_probs( void );
void rnd_sym_probs(double prob_syms[]);

//	------- FILE OPS -------
void load_model(char* filename);
int Read_Sequences(char* filename);

//	------ CALCULATIONS -------
void rescale_alphas(int col);
void rescale_betas(int col);
double alpha_F(int* symbol_array, int seq_length);
double beta_I(int* symbol_array, int seq_length);
void compute_gamxi(int* symbol_array, int seq_length);
double a_numer(int i, int j, int seq_length);						// also b_denom
double a_denom(int i, int seq_length);
double b_numer(int i, int sym, int *symbol_array, int seq_length);
double b_denom(int i, int seq_length);
double set_ab_counts(int* string_array,int seq_length);
double reestimate(void);																		// returns sum of changes
double test(int* string, int seq_length);										// test one string 

//	------ VITERBI ROUTINES -------

void Viterbi( int index);									// compute MLE state sequence for string[index] using	Viterbi algorithm from Rabiner
double qprop(int state, int length);			// return proportion of qstar that have the value state

 //	------ HMM ROUTINES -------

void Set_Size(void);																			// set globals max_symbols and max_states
void Init_Random_HMM(void);																// initialize random model
void Init_Prebuilt_HMM(char* filename);                   // initialize pre-built model
void Init_Saved_HMM(int isave);														// set max_states, max_symbols and model parameters from file
void Init_Saved_HMMXC(int isave);													// same for saved XC 
void batch_train(int seq_length, double min_delta_psum);	// estimate model
double batch_test(char *filename, int isave);	     				// test using file
void check_save(void);																		// check whether we should save the model
void check_save_XC(double sumpr);													// same for XC model
void show_probs(void);                         						// display model probabilities
void dump_model(char *fname);           									// dumps the current model

//	------ WEIS ROUTINES -------

long Julian (char *s);																				// Converts dates of form YYMMDD to Julian serial beginning 1 Jan 1900 
void Read_Ex_WEIS(char *SourceCode, char *TargetCode);				// read a WEIS ME file while filtering non-BCOW events
void Read_Dyadic_WEIS(char *SourceCode, char *TargetCode);		// read a dyadic WEISDY file
void Init_Template_HMM(char *sFile);													// Reads a set of dates from sFile; moves SEQ_SIZE events into strings, then estimates a model
void Estimate_HMM(void);																			// MC estimation from existing set of strings
void Next_Month (char *sMonth);																// increment the month string
void Compute_Running_Prob (char *SourceCode, char *TargetCode, char ctype );			// computes alpha for 100 events prior to the end of each month


//	------ XCORR ROUTINES -------

double XCorr(double x[], double y[], int xa, int ya, int max);	// Cross-correlation function
void Read_Indicator(void);																		// read the indicator array											
int Get_Index(double r);
double Sum_XCorr(void);																				// compute the sum of the target cross-correlations

//__________________________________________________________
// Globals
  
// Monte-Carlo experiments
state			aStates[MAX_STATES];   						// matrix of model parameters
state			aSaveSt[MAX_STATES][MAX_SAVE];   	// array for saving models
double		aSaveAlpha[MAX_SAVE];   					// sum of alpha of saved models
double		curAlpha;													// tot_alpha from last computation

state			aSaveStXC[MAX_STATES][MAX_SAVE];  // array for saving XC models
double		aSaveXC[MAX_SAVE];   							// sum xcorr of saved models

// from class HMM:

int		max_states;                   			// number of states in the model
int		max_symbols;                  			// symbol emissions per state
int		trellis_width;                			// cols allocated in trellis =num_symbols-1
int		num_strings;           							// number of training/testing strings
int		strings[MAX_STRINGS][MAX_LENGTH];   // matrix of training/testing strings
int		string_len[MAX_STRINGS];		    		// sequence length
char	string_name[MAX_STRINGS][NAME_LENGTH];		// sequence name
double					prob_seq;								// sequence probability given current model (computed in alpha_F)

arTrellState 		alpha;                  // matrix for alpha trellis
arTrellState 		beta;                   // matrix for beta trellis
arTrellState 		gammaF;                  // matrix for gamma trellis
arTrell				  scaling_factors;        // array of alpha and beta scaling factors
arTrellState 		xi_prev;             		// matrix of xi prev transition probs
arTrellState 		xi_next;             		// matrix of xi next transition probs
arTrellState 		xi_recur;            		// matrix of xi recurrence probs
arState 				a_numer_sum_recur;      // array of numerators for a_ij's
arState 				a_numer_sum_next;       // array of numerators for a_ij's
arState 				a_numer_sum_prev;       // array of numerators for a_ij's
arState 				a_denom_sum;      			// array of denomonators for a_ij's
arStateSymb 		b_numer_sum;      			// array of numerators for b_ij's
arState 				b_denom_sum;       			// array of denominators for b_ij's

arTrellState 		delta;                  // matrix for delta trellis (Viterbi alg.)
int 						psi[MAX_TRELLIS][MAX_STATES] ; // same for gamma	
int 						qstar[MAX_TRELLIS] ;		// holds the optimal state sequence

double					indic[MAX_MONTHS];				// target indicator for xcorr
double					prstate[MAX_MONTHS];			// percentage state indicator for xcorr
int							nMonth;	

long						aWEIS_Date[MAX_WEIS];		// dates of WEIS events
int							aWEIS_Evt[MAX_WEIS];		// WEIS event sequence
int							numWEIS;								// number of WEIS events
int			Month[] = {0,0,31,59,90,120,151,181,212,243,273,304,334};		// used for Julian computation: # days prior to Month[n] JAN=1
